#!/usr/bin/env python
""" Find contigs with deletions in contig composition summary.
"""

from Bio import SeqIO
from Bio.Alphabet import generic_dna
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from vamp.utils import ContigComposition
import Bio.AlignIO
import Bio.SeqIO
import argparse
import os
import re
import sys
import vamp.utils
from vamp import __version__ as version

__author__ = "Lance Parsons"
__author_email__ = "lparsons@princeton.edu"
__copyright__ = "Copyright 2013, 2014 Lance Parsons"
__license__ = ("BSD 2-Clause License "
               "http://www.opensource.org/licenses/BSD-2-Clause")


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description=''' Find contigs with deletions from the
        contig composition file output from compare_genomes.py''')
    parser.add_argument('contig_composition', help="Contig composition file "
                        "output from compare_genomes.py")
    parser.add_argument('aligned_fasta', help="Aligned FASTA file")
    parser.add_argument('contigs_fasta', help="Contigs FASTA file")
    parser.add_argument('-o', '--output_dir', default=None,
                        help='Directory to store output files, default is '
                        'aligned_fasta directory')
    parser.add_argument('-q', '--quiet', action='store_true', default=False,
                        help='Quiet, replace all deletions found, no prompts')
    parser.add_argument('-v', '--verbose', action='store_true',
                        default=False, help='verbose output')
    parser.add_argument('--version', action='version',
                        version='%(prog)s version ' + version)
    args = parser.parse_args(argv[1:])

    contig_composition_file = args.contig_composition
    aligned_fasta_input = args.aligned_fasta
    contigs_fasta = args.contigs_fasta
    (output_dir, output_basename) = os.path.split(aligned_fasta_input)
    (output_basename, dummy_ext) = os.path.splitext(output_basename)
    if args.output_dir:
        output_dir = args.output_dir

    # Read contigs file
    fasta_dict = Bio.SeqIO.index(contigs_fasta, 'fasta')
    contigs = []
    sequence_source_id = None
    print "\nContig Composition Summary"
    print ContigComposition.tab_headings()
    for contig in vamp.utils.read_contig_composition_summary(
            contig_composition_file):
        contigs.append(contig)
        print contig.to_tab()
        sequence_source_id = contig['seq']
    fasta_output_filename = os.path.join(
        output_dir, "%s_deletions_added.fasta" % sequence_source_id)

    # Read Alignment
    alignment = Bio.AlignIO.read(aligned_fasta_input, 'fasta')
    original_sequence_alignment = None
    for idx, seq in enumerate(alignment):
        if seq.id == sequence_source_id:
            original_sequence_alignment = seq
    assembled_sequence_str = re.sub('-', '',
                                    str(original_sequence_alignment.seq))
    if args.verbose:
        print("\nOriginal Sequence of %s:\n%s\n" %
              (original_sequence_alignment.id, assembled_sequence_str))
    assembled_sequence = SeqRecord(Seq(assembled_sequence_str, generic_dna),
                                   id=original_sequence_alignment.id)

    # Find deletions
    replacements = vamp.utils.find_deletions(contigs, True)

    # Allow user to select them
    selected_replacements = []
    for (contigs_to_be_replaced, replacement) in replacements:
        print ("\nContig %s:%s-%s is missing" %
               (replacement['contig'],
                replacement['contig_start'],
                replacement['contig_end']))
        if not contigs_to_be_replaced:
            print ("\tinsert\n\t\t%s\n\tat %s:%s" %
                   (replacement.to_tab(),
                    replacement['seq'],
                    replacement['start']))
        else:
            print ("\treplace")
            for idx in contigs_to_be_replaced:
                contig_to_be_replaced = contigs[idx]
                print ("\t\t%s" % contig_to_be_replaced.to_tab())
            print "\twith\n\t\t%s" % replacement.to_tab()
        if (args.quiet or confirm("\t? ")):
            selected_replacements.append(replacement)

    # Generate updated sequence
    updated_sequence = vamp.utils.update_sequence_with_replacements(
        assembled_sequence, selected_replacements, fasta_dict)
    if args.verbose:
        print ("\nUpdated Sequence of %s:\n%s\n" %
               (original_sequence_alignment.id, updated_sequence.seq))
    SeqIO.write(updated_sequence, fasta_output_filename, "fasta")

    # TODO Generate updated alignment

    # TODO Generate updated contig composition summary


def confirm(prompt=None, resp=False):
    """prompts for yes or no response from the user. Returns True for yes and
    False for no.

    'resp' should be set to the default value assumed by the caller when
    user simply types ENTER.

    >>> confirm(prompt='Create Directory?', resp=True)
    Create Directory? [y]|n:
    True
    >>> confirm(prompt='Create Directory?', resp=False)
    Create Directory? [n]|y:
    False
    >>> confirm(prompt='Create Directory?', resp=False)
    Create Directory? [n]|y: y
    True

    """

    if prompt is None:
        prompt = 'Confirm'

    if resp:
        prompt = '%s [%s]|%s: ' % (prompt, 'y', 'n')
    else:
        prompt = '%s [%s]|%s: ' % (prompt, 'n', 'y')

    while True:
        ans = raw_input(prompt)
        if not ans:
            return resp
        if ans not in ['y', 'Y', 'n', 'N']:
            print 'please enter y or n.'
            continue
        if ans == 'y' or ans == 'Y':
            return True
        if ans == 'n' or ans == 'N':
            return False

if __name__ == '__main__':
    sys.exit(main())
