#!/usr/bin/env python

'''
Extracts coding sequences (cds) regions from fasta reference and gff file and
translates them into amino acid sequence, output in FASTA format to STDOUT.

Errors during translation are output to STDERR. Genes with translation errors
are not printed.
'''
import argparse
from collections import defaultdict
import os
import sys

from Bio import SeqIO
from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Data import CodonTable
from pybedtools import BedTool

from vamp import __version__ as version

__author__ = "Lance Parsons"
__author_email__ = "lparsons@princeton.edu"
__copyright__ = "Copyright 2013, 2014 Lance Parsons"
__license__ = "BSD 2-Clause License " \
    "http://www.opensource.org/licenses/BSD-2-Clause"


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(
        add_help=False, description=''' Extracts the coding sequences
        (cds) regions from a fasta reference and gff file and translates them
        into amino acid sequences, output in FASTA format to STDOUT''')

    # Required arguments
    req_arguments = parser.add_argument_group('Required Arguments')
    req_arguments.add_argument('gff_file', help="GFF file containing "
                               "CDS records to be translated")
    req_arguments.add_argument('fasta_file', help="FASTA file "
                               "containing the nucleotide sequences "
                               "referenced in the GFF file")

    # Optional arguments
    opt_arguments = parser.add_argument_group('Optional Arguments')
    opt_arguments.add_argument(
        '--notrans', action='store_true', default=False,
        help='Do not translate to amino acid sequence, output DNA')
    opt_arguments.add_argument(
        '-i', '--idattr', default='gene_id',
        help='GFF attribute to use as gene ID. Features with the same ID will '
        'be considered parts of the same gene. The default "%(default)s" is '
        'suitable for GTF files.')
    opt_arguments.add_argument(
        '-t', '--featuretype', default=None, action='append',
        help='GFF feature type(s) (3rd column) to be used. Specify the option '
        'multiple times for multiple feature types. The default is "CDS" for '
        'GFF files and "CDS" and "stop_codon" for GTF files.')
    opt_arguments.add_argument(
        '--table', default=1, help='NCBI Translation table to '
        'use when translating DNA (see http://www.ncbi.nlm.nih.'
        'gov/Taxonomy/Utils/wprintgc.cgi). Default: %(default)s.')
    opt_arguments.add_argument('-v', '--verbose', action='store_true',
                               default=False, help='verbose output')
    opt_arguments.add_argument('--version', action='version',
                               version='%(prog)s version ' + version)
    opt_arguments.add_argument('-h', '--help', action='help',
                               help='show this help message and exit')

    args = parser.parse_args(argv[1:])

    # Get Arguments
    gff_filename = args.gff_file
    fasta_filename = args.fasta_file

    # Read GFF file
    basename, extension = os.path.splitext(gff_filename)
    is_gtf = False
    if extension.lower() == '.gtf':
        is_gtf = True
    if args.featuretype is None:
        featuretypes = ['CDS']
        if is_gtf:
            featuretypes.append('stop_codon')
    else:
        featuretypes = args.featuretype
    gff_file = BedTool(gff_filename)
    if gff_file.file_type != 'gff':
        parser.error('Gene file format not recognized, must be GFF or GTF')
    cds_regions = gff_file.filter(lambda b: b.fields[2] in
                                  featuretypes)
    genes = defaultdict(list)
    for feature in cds_regions:
        gene_id = feature.attrs.get(args.idattr)
        if gene_id is not None:
            genes[gene_id].append(feature)
    # TODO Store gene_name to use in fasta description
    # TODO Store start position (and chromosome order) and use to sort

    # Read Fasta File
    with open(fasta_filename) as in_handle:
        ref_recs = SeqIO.to_dict(SeqIO.parse(in_handle, "fasta"))

    for gene, cds in genes.iteritems():
        seq = Seq("", IUPAC.unambiguous_dna)
        for c in cds:
            ref_seq = ref_recs.get(c.chrom)
            cds_seq = ref_seq[c.start:c.end]
            if c.strand == '-':
                cds_seq = cds_seq.reverse_complement()
            seq = seq + cds_seq.seq

        # Attempt to translate as complete CDS, report errors
        protein_seq = None
        try:
            protein_seq = seq.translate(table=args.table, cds=True)
        except CodonTable.TranslationError as e:
            sys.stderr.write("Error found in gene '%s': %s\n" %
                             (gene, str(e)))

        # Attempt to translate without requiring complete CDS
        if protein_seq is None:
            try:
                protein_seq = seq.translate(table=args.table, cds=False)
            except Exception as e:
                sys.stderr.write("Unable to translate gene '%s': %s\n" %
                                 (gene, str(e)))
        # Print output
        output_record = None
        if args.notrans:
            output_record = SeqRecord(seq, id=gene, description='')
        else:
            output_record = SeqRecord(protein_seq, id=gene, description='')
        if output_record is not None:
            print output_record.format("fasta"),

    return 0


if __name__ == '__main__':
    sys.exit(main())
