#!/usr/bin/env python
"""
Summarize the differences between sequences in an aligned FASTA file.

This script will output summarize the differences between sequences in an
aligned FASTA file.

Usage::

    summarize_alignments.py aligned_fasta reference_sequence [-h,--help]
        [-v,--verbose] [--version]

"""

"""
Copyright 2012, 2013, 2014 Lance Parsons <lparsons@princeton.edu>. All rights
reserved.

BSD 2-Clause License http://www.opensource.org/licenses/BSD-2-Clause
"""

from Bio import AlignIO
import optparse
import os
import sys
import time
import traceback


def main():
    """
    Runs summary_of_alignment function on input files from the command line.
    """
    global options, args

    # Read multi-fasta file
    multi_align_file = args[0]
    reference_sequence_id = args[1]
    alignment = AlignIO.read(open(multi_align_file), "fasta")
    print ("There are %i sequences in this alignment" % len(alignment))

    # Find reference sequence
    ref_found = False
    for s in alignment:
        print "\t%i: %s" % (s.id),
        if (s.id == reference_sequence_id):
            print " (Reference)",
            ref_found = True
        print "\n",
    print ("The alignment is %i bases long" % len(alignment[0, :]))
    if not ref_found:
        sys.stderr.write(
            "Reference sequence id '%s' not found in fasta file '%s'" %
            (reference_sequence_id, multi_align_file))
        return 1

    mismatch_summaries = summary_of_alignment(alignment, reference_sequence_id)
    for seq_id, mismatch_summary in mismatch_summaries.iteritems():
        print "\nSequence: %s" % seq_id
        print "Matches:       %i" % mismatch_summary['match_count']
        print "Mismatches:    %i" % mismatch_summary['mismatch_count']
        print "Change Events: %i" % mismatch_summary['noncontig_count']


def summary_of_alignment(alignment, reference_sequence_id):
    '''
    Summarizes changes in given alignment

    Args:
        alignment (Bio.AlignIO object): Alignment object
        reference_index (int):  index of the reference sequence in alignment
            (default is 1)

    Returns:
        dictionary: A dctionary with a key for each non-reference sequence
            in the alignment

        Each entry is another dictionary with the following keys:

        * *match_count*: The number of matching bases
        * *mismatch_count*: The number of mismatching bases, including indels
        * *mismatches*: list of mismatches by base: `RefBase(RefPos)NewBase`
        * *contiguous_change_count*: the number of contiguous change
                "events"
    '''
    mismatch_summaries = {}
    reference_index = None
    for i in range(len(alignment)):
        if (alignment[i].id != reference_sequence_id):
            mismatch_summaries[alignment[i].id] = dict(match_count=0,
                                                       mismatch_count=0,
                                                       mismatches=[],
                                                       noncontig_count=0)
        else:
            reference_index = i

    for sequence in alignment:
        if sequence.id == reference_sequence_id:
            continue
        event_list = []
        event = dict(pos=0, ref='', alt='', align_pos=0, type=None, snp=False,
                     insertion=False, deletion=False)
        mismatch_event_count = 0
        refposition = 0
        previous_position = -2
        for position in range(len(alignment[reference_index])):
            refbase = alignment[reference_index][position]
            altbase = sequence[position]
            if (refbase != '-'):
                refposition += 1
            if (refbase != altbase):
                mismatch_summaries[sequence.id]['mismatch_count'] += 1
                mismatch_summaries[sequence.id]['mismatches'].append(
                    dict(pos=refposition, ref=refbase, alt=altbase))
                if refposition > previous_position + 1:
                    mismatch_event_count += 1
                    if event['pos'] != 0:
                        # Append previous event to list
                        variation = parse_event(
                            event, alignment[reference_index], sequence)
                        event_list.append(variation)
                    event = dict(pos=0, ref='', alt='', align_pos=0, type=None,
                                 snp=False, insertion=False, deletion=False)
                    event['pos'] = refposition
                    event['ref'] = refbase
                    event['alt'] = altbase
                    event['align_pos'] = position
                else:
                    event['ref'] += refbase
                    event['alt'] += altbase
                # Set flags used to determine type
                if refbase == '-':
                    event['insertion'] = True
                elif altbase == '-':
                    event['deletion'] = True
                else:
                    event['snp'] = True
                previous_position = refposition
        else:
            mismatch_summaries[sequence.id]['match_count'] += 1
        # Summarize mismatch info
        mismatch_summaries[sequence.id]['noncontig_count'] = len(event_list)
        mismatch_summaries[sequence.id]['noncontig_list'] = event_list
        mismatch_summaries[sequence.id]['noncontigs_string'] =\
            mismatch_string(event_list)
        mismatch_summaries[sequence.id]['mismatches_string'] =\
            mismatch_string(mismatch_summaries[sequence.id]['mismatches'])
    return(mismatch_summaries)


def parse_event(event, reference_sequence, alternate_sequence):
    """ Parse an event (sequence of differences) for VCF output.

    Parse a simple event with reference_position, reference_base, and new_base
    and determine the type and add padding if necessary (for VCF compatibility)

    Args:
        event (dictonary): An event has at least a position (pos), reference
            genotype (ref), and aternate genotype (alt). May also have a flag
            indicating if it is a snp (snp).
        reference_sequence (str): The complete reference sequence
        alternate_sequence (str): The complete alternate sequence

    Returns:
        dictonary: An event with additional padding to the start of the variant
            and an added type attribute, for VCF compatibility
    """
    refgaps = event.get('ref').count('-')
    altgaps = event.get('alt').count('-')

    if refgaps > 0 or altgaps > 0:
        if event['pos'] > 1:  # Add padding base to start of variant
            event['pos'] -= 1
            event['ref'] = '%s%s' % (
                reference_sequence[event['align_pos'] - 1], event['ref'])
            event['alt'] = '%s%s' % (
                alternate_sequence[event['align_pos'] - 1], event['alt'])

    if event.get('snp'):
        if (len(event.get('ref')) == 1 and len(event.get('alt')) == 1):
            event['type'] = 'SNP'
        else:
            if refgaps == 0 and altgaps == 0:
                event['type'] = 'MNP'
            else:
                event['type'] = 'COMPLEX'
    else:
        if refgaps > 0 or altgaps > 0:
            event['type'] = 'INDEL'
    return event


def mismatch_string(mismatches):
    """ Generate a string from a list of mismatches.

    Args:
        mismatches (list): A list of mismatches. A mismatch is a dictionary
            with a position (pos), reference genotype (ref), and alternate
            genotype (alt).

    Returns:
        string: A comma separated string of the mismatches
    """

    mismatch_string_list = list()
    for m in mismatches:
        mismatch_string_list.append(
            "%s(%i)%s" % (m.get("ref"), (m.get("pos") + 1), m.get("alt")))
    return ', '.join(mismatch_string_list)


if __name__ == '__main__':
    try:
        start_time = time.time()
        parser = optparse.OptionParser(
            formatter=optparse.TitledHelpFormatter(),
            usage=globals()['__doc__'], version='$Id$')
        parser.add_option(
            '-v', '--verbose', action='store_true', default=False,
            help='verbose output')
        (options, args) = parser.parse_args()
        if len(args) < 2:
            parser.error('Please specify an aligned fasta file and a '
                         'reference sequence id')
        if options.verbose:
            print time.asctime()
        main()
        if options.verbose:
            print time.asctime()
            print 'TOTAL TIME IN MINUTES:',
            print (time.time() - start_time) / 60.0
        sys.exit(0)
    except KeyboardInterrupt, e:  # Ctrl-C
        raise e
    except SystemExit, e:  # sys.exit()
        raise e
    except Exception, e:
        print 'ERROR, UNEXPECTED EXCEPTION'
        print str(e)
        traceback.print_exc()
        os._exit(1)
