"""
Utilities for working with multiple sequence alignments and MAF objects

Copyright 2012, 2013, 2014 Lance Parsons <lparsons@princeton.edu>
All rights reserved.

BSD 2-Clause License http://www.opensource.org/licenses/BSD-2-Clause
"""

from Bio.Alphabet import generic_dna
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from galaxy import maf_utilities
import Bio.Align
import Bio.SeqIO
import bx.align.maf
import copy
import os
import pybedtools
import re
import seq_utils.utils
import sys


def get_vamp_home():
    """ Return the directory where the VAMP module is installed """
    VAMP_HOME = os.path.abspath(os.path.join(os.path.dirname(
        sys.modules[__name__].__file__), '..'))
    return VAMP_HOME


def get_sequence_net_alignment(maf_filename, reference_species, sequence_id,
                               species, verbose=False):
    """
    Return the alignment created by stitching MAF blocks together

    Stitches MAF blocks together along an entire reference sequence (including
    gaps).  For regions covered by more than one block, the highest scoring
    block is used.

    Args:
        maf_filename: The filename of the MAF file.
        reference_species: The name species used as the reference.
        sequence_id: The sequence_id used as the reference. The format of
            sequence names in the MAF file is assumed to be
            'species.sequence_id' (e.g. 'scerevisiae.chrI')
        species: A list of the species names to be returned
        verbose (bool, optional): If True, print debug information (default:
            False)

    Returns:
        tuple: A tuple containing a Bio.Align.MutlipleSeqAlignment object  and
        a list of intervals.  The multiple sequence alignments contains each
        the alignment of each species from the MAF file created by stitching
        blocks together based on the specified reference sequence The list of
        intervals is relative to the alignment that indicate the MAF block,
        block start, and block end of the source of that piece of the
        alignment.
    """

    maf_file = open(maf_filename, 'rU')

    # Get reference.sequence_id length
    ref_seq_length = get_sequence_length_from_maf(maf_file, reference_species,
                                                  sequence_id)
    if ref_seq_length is None:
        raise Exception('Source "%s.%s" not found in maf file: "%s"' %
                        (reference_species, sequence_id, maf_filename))
    if verbose:
        print ("Size of %s.%s: %i bp" %
               (reference_species, sequence_id, ref_seq_length))

    # Create alignment with full sequence from fasta for reference.sequence_id,
    # use N for all other species in alignment
    sequence_list = []
    for spec in species:
        seq = Seq('N' * ref_seq_length, generic_dna)
        src = "%s.%s" % (spec, sequence_id)
        sequence_list.append(SeqRecord(seq, id=src))
    alignment = Bio.Align.MultipleSeqAlignment(sequence_list)
    if verbose:
        print alignment

    # Get all blocks with this species.sequence_id
    # Sort block by score, lowest first
    # index maf for use here, need to remove index_file when finished
    index, index_filename = maf_utilities.open_or_build_maf_index(
        maf_file.name, '%s.idx' % maf_file.name, species=[reference_species])
    if index is None:
        sys.stderr.write("Your MAF file appears to be malformed.\n")

    # Order blocks overlapping this position by score, lowest first
    blocks = []
    source = '%s.%s' % (reference_species, sequence_id)
    if verbose:
        print "Reordering MAF Blocks"
    for block, idx, offset in \
            index.get_as_iterator_with_index_and_offset(source, 0,
                                                        ref_seq_length):
        score = float(block.score)
        for i in range(0, len(blocks)):
            if score < blocks[i][0]:
                blocks.insert(i, (score, idx, offset))
                if verbose:
                    print "Adding block %s" % block.attributes['label']
                break
        else:
            blocks.append((score, idx, offset))
            if verbose:
                print "Adding block %s" % block.attributes['label']

    # Replace section of alignment with alignment from the MAF block
    interval_list = []
    for block_dict in blocks:
        block = block_dict[1].get_at_offset(block_dict[2])
        if verbose:
            print "Analyzing maf block %s" % block.attributes['label']
        for c in block.components:
            # TODO Refactor to function and only allow single period
            (spec, dummy_seq_id) = c.src.split('.', 1)
            if verbose:
                print "    Component src %s" % c.src
            if spec == reference_species:
                # Replace alignment and calculate difference in length
                original_alignment_length = len(alignment[0])
                alignment, new_interval = replace_alignment_with_block(
                    alignment, block, reference_species, sequence_id, verbose)
                interval_adjustment = (len(alignment[0]) -
                                       original_alignment_length)
                if verbose:
                    print "interval_adjustment = %s" % interval_adjustment
                # Generate new interval and add to list
                interval_list_string = ''
                for interval in interval_list:  # [0:len(interval_list) - 1]:
                    trimmed_intervals = subtract_intervals(interval,
                                                           new_interval)
                    for trimmed_interval in trimmed_intervals:
                        block_start_adjustment = (trimmed_interval.start -
                                                  interval.start)
                        block_end_adjustment = (trimmed_interval.end -
                                                interval.end)
                        block_updated_interval = \
                            pybedtools.create_interval_from_list([
                                trimmed_interval.chrom,
                                "maf_net",
                                "alignment_block",
                                str(trimmed_interval.start + 1),
                                str(trimmed_interval.end),
                                ".",
                                '+',
                                ".",
                                "maf_block=%s;block_start=%i;block_end=%i" %
                                (trimmed_interval.attrs['maf_block'],
                                 long(trimmed_interval.attrs['block_start']) +
                                    block_start_adjustment,
                                 long(trimmed_interval.attrs['block_end']) +
                                    block_end_adjustment)
                            ])
                        interval_list_string += (str(block_updated_interval) +
                                                 "\n")
                interval_list_string += str(new_interval) + "\n"
                interval_list = pybedtools.BedTool(str(interval_list_string),
                                                   from_string=True)
                # For each interval, update start and end by interval
                # adjustment to account for gaps
                interval_list_string = ''
                for interval in interval_list:
                    if interval.start >= new_interval.end:
                        interval.start = interval.start + interval_adjustment
                    if interval.end >= new_interval.end:
                        interval.end = interval.end + interval_adjustment
                    interval_list_string += str(interval) + "\n"
                interval_list = pybedtools.BedTool(str(interval_list_string),
                                                   from_string=True)

                if verbose:
                    for seq in alignment:
                        print("%s: Number of bases %i" %
                              (seq.id, len(re.sub('-', '', str(seq.seq)))))

    # Return alignment
    os.unlink(index_filename)
    maf_file.close()
    return alignment, interval_list


def subtract_intervals(interval1, interval2):
    """
    Subtract two pybedtools intervals, return list of resulting intervals

    Args:
        interval1: A pybedtools interval
        interval2: A pybedtools interval to subtract from interval1

    Returns:
        list: A list of pybedtools intervals that contain the
            region(s) of interval1 that are not overlapped by interval2
    """
    results = [copy.deepcopy(interval1)]
    if (interval2.end <= interval1.end and
            interval2.end > interval1.start):
        if (interval2.start >= interval1.start and
                interval2.start < interval1.end):
            results.append(copy.deepcopy(interval1))
            results[0].end = interval2.start
            results[1].start = interval2.end
        else:
            results[0].start = interval2.end
    elif (interval2.start >= interval1.start and
            interval2.start < interval1.end):
        results[0].end = interval2.start
    return results


def replace_alignment_with_block(alignment, block, reference_species,
                                 sequence_id, verbose=False):
    """
    Update the multiple sequence alignment with the specified MAF block

    Use the MAF block alignment to replace the appropriate section of the given
    multiple sequence alignment by using the specified reference species and
    sequence as guide

    Args:
        alignment (MultipleSequenceAlignment):
        block (maf block): MAF block
        reference_species (str): The name species used as the reference.
        sequence_id (str): The sequence_id used as the reference. The format of
            sequence names in the MAF file is assumed to be
            'species.sequence_id' (e.g. 'scerevisiae.chrI')
        verbose (bool, optional): If True, print debug information (default:
            False)

    Returns:
        tuple: The updated alignment and a Pybedtools interval of the section
        of the alignment that was replaced.  The interval contains the
        following attributes: `maf_block; block_start; block_end` which
        indicate the MAF block label and start and end position on the block
        used in the replacement
    """

    src = '%s.%s' % (reference_species, sequence_id)
    ref_component = block.get_component_by_src(src)
    if verbose:
        print "    Reference component %s" % ref_component
    # Orient block so reference is forward strand
    if ref_component.strand == '-':
        block = block.reverse_complement()
        ref_component = block.get_component_by_src(src)
        if verbose:
            print "  Reference in reverse orientation"
    ref_component_length_withgaps = len(ref_component.text)
    ref_align_seq = seq_utils.utils.get_seq_by_id_from_list(alignment, src)
    if verbose:
        print "Ref align seq %s" % ref_align_seq
    if ref_align_seq is None:
        return alignment
    find_start = ref_component.start
    find_end = ref_component.start + ref_component.size
    if verbose:
        print (str(ref_align_seq.seq), find_start, find_end)
    alignment_replace_start, alignment_replace_end = (
        seq_utils.utils.convert_interval_nongapped_to_gapped(
            str(ref_align_seq.seq), find_start, find_end))
    if verbose:
        print ('%s - start: %i - end: %i' %
               (ref_component.src, find_start, find_end))
        print ('%s - length (inc gaps): %i' %
               (ref_component.src, ref_component_length_withgaps))
        print ('%s - length (no gaps): %i' %
               (ref_component.src, ref_component.size))
        print ('%s - Replace Start: %i - End: %i' %
               (ref_component.src, alignment_replace_start,
                alignment_replace_end))
        print ('%s - Size of text to be replaced: %i' %
               (ref_component.src,
                (alignment_replace_end - alignment_replace_start)))
        print 'Length of alignment before replacement: %i' % len(alignment[0])

    # Replace Sequences one by and one and construct new alignment
    sequence_list = []
    for sequence in alignment:
        (seq_spec, dummy_seq_id) = sequence.id.split('.', 1)
        replacement_text = ('N' * (ref_component_length_withgaps))
        for c in block.components:
            (c_spec, dummy_c_id) = c.src.split('.', 1)
            if c_spec == seq_spec:
                replacement_text = c.text
                if verbose:
                    print ("%s: Replacement text: %s" %
                           (c.src, replacement_text))
                    print ("%s: Length of block replacement text: %i" %
                           (c.src, len(replacement_text)))
                    print ('%s: Number of bases in replacement text: %i' %
                           (c.src, len(re.sub('-', '', replacement_text))))
        new_sequence_text = str(sequence.seq)
        new_sequence = Seq(new_sequence_text[0:alignment_replace_start] +
                           replacement_text +
                           new_sequence_text[alignment_replace_end:],
                           generic_dna)
        sequence_list.append(SeqRecord(new_sequence, id=sequence.id))
        if verbose:
            new_seq_piece = new_sequence_text[
                alignment_replace_start:alignment_replace_end]
            print ('%s: Number of bases to be replaced: %i' %
                   (sequence.id, len(re.sub('-', '', new_seq_piece))))
    new_alignment = Bio.Align.MultipleSeqAlignment(sequence_list)

    block_start = 0 + (find_start - find_start)
    block_end = long(len(replacement_text)) - (find_end - find_end)
    new_interval = pybedtools.create_interval_from_list([
        sequence_id,
        "maf_net",
        "alignment_block",
        str(alignment_replace_start + 1),
        str(alignment_replace_end),
        ".",
        '+',
        ".",
        "maf_block=%s;block_start=%i;block_end=%i" %
        (block.attributes['label'], block_start, block_end)
    ])

    if verbose:
        print ('Length of alignment after replacement: %i' %
               len(new_alignment[0]))
        print block.attributes['label']
    return new_alignment, new_interval


def get_sequence_length_from_maf(maf_file, reference_species, sequence_id):
    """
    Return length of the reference_species.sequence_id

    Args:
        maf_filename: The filename of the MAF file.
        reference_species: The name species used as the reference.
        sequence_id: The sequence_id used as the reference. The format of
            sequence names in the MAF file is assumed to be
            'species.sequence_id' (e.g. 'scerevisiae.chrI')

    Returns:
        integer: The length of the specified sequence in the first component
            containing that sequence in the MAF file, or None if no matching
            componenets were found in the MAF file.
    """
    for block in bx.align.maf.Reader(maf_file):
        for component in block.components:
            if component.src == '%s.%s' % (reference_species, sequence_id):
                return component.get_src_size()
    return None


def verify_maf_fasta(maf_filename, reference_species, fasta_filename,
                     verbose=False):
    """
    Verify the consistency between the sequence in a MAF and a FASTA file

    Checks all compenents in all blocks of the MAF file for the specified
    species and checks that the sequence matches that in the FASTA file.

    Args:
        maf_filename (string): The name of the MAF file.
        reference_species (string): The species to select from the MAF file.
        fasta_filename (string): The name of the FASTA file to check against.
        verbose (bool, optional): If true, output additional debugging info
            (default is False).

    Returns:
        None: Prints to STDOUT if there is a mismatch.
    """

    maf_file = open(maf_filename, 'rU')
    fasta_file = open(fasta_filename, 'rU')
    record_dict = Bio.SeqIO.to_dict(Bio.SeqIO.parse(fasta_file, 'fasta'))
    if verbose:
        print record_dict.keys()
    for block in bx.align.maf.Reader(maf_file):
        for component in block.components:
            (species, seq_id) = component.src.split('.', 1)
            if species == reference_species:
                if component.strand == '-':
                    block = block.reverse_complement()
                    component = block.get_component_by_src(component.src)
                start = component.start
                end = component.start + component.size
                if verbose:
                    print ('Checking block %s: %s:%i-%i' %
                           (block.attributes['label'], seq_id, start, end))
                fasta_sequence = str(record_dict[seq_id][start:end].seq)
                maf_sequence = re.sub('-', '', component.text)
                if fasta_sequence != maf_sequence:
                    print "No Match!"
                    print ">FASTA.%s\n%s" % (seq_id, fasta_sequence)
                    print ">MAF.%s\n%s" % (seq_id, maf_sequence)
    fasta_file.close()
    maf_file.close()


def get_block_by_label(maf_filename, label):
    """ Return the MAF block with the specified label

    Args:
        maf_filename (string): The name of the MAF file.
        label (string): The label of the block in the MAF
            file to search for.

    Returns:
        block: The first block found in the MAF file that has
            the given label
    """

    selected_block = None
    maf_file = open(maf_filename, 'rU')
    for block in bx.align.maf.Reader(maf_file):
        if block.attributes['label'] == label:
            selected_block = block
            break
    return selected_block


def summarize_contig_composition(interval_list, src_tag, start_tag, end_tag,
                                 strand_tag, source_size_tag):
    """
    Summarize the contig composition of a stitched MAF file.

    Args:
        interval_list (list): A list of Pybedtools interval objects
        src_tag (string): The attribute containing the contig name
        start_tag (string): The attribute containing the start position in the
            contig
        end_tag (string): The attribute containing the end position in the
            contig
        strand_tag (string): The attribute containing the strand
        source_size_tag (string): The attribute containing the contig size

    Returns:
        list: A list of dictionaries with the following keys: (seq, start, end,
            contig, contig_start, contig_end, strand, contig_size)
    """
    section_list = []
    for interval in interval_list:
        strand = interval.attrs[strand_tag]
        start = interval.attrs[start_tag]
        end = interval.attrs[end_tag]
        section = {'seq': interval.chrom,
                   'start': interval.start,
                   'end': interval.end,
                   'contig': interval.attrs[src_tag],
                   'contig_start': start,
                   'contig_end': end,
                   'strand': strand,
                   'contig_size': interval.attrs[source_size_tag]}
        section_list.append(section)
    return section_list


def read_contig_composition_summary(filename):
    """
    Generator that reads a contig composition summary file and returns
        attributes.

    Args:
        filename (string): The name of contig composition summary file
            as output by compare_genomes.py.

    Yields:
        ContigComposition: A :class:`ContigComposition` object for each line in
            the contig composition summary file.
    """
    linenum = 0
    fh = open(filename, 'rb')
    for line in fh:
        linenum += 1
        if (line[0] != '#') and (linenum != 1):
            line = line.strip()
            fields = line.split('\t')
            section = ContigComposition({'seq': fields[0],
                                         'start': fields[1],
                                         'end': fields[2],
                                         'contig': fields[3],
                                         'contig_start': fields[4],
                                         'contig_end': fields[5],
                                         'strand': fields[6],
                                         'contig_size': fields[7]})
            yield section


def find_deletions(contig_composition_list, verbose=False):
    """
    Find contigs with deletions.

    From a contig composition list, find contigs that have deleted sections.
    When a contig has deleted sections, the pieces of the contig may be
    replaced by a contiguous section.  This function returns tuples containing
    the indices of the contigs pieces to be replaced along with a replacement
    :class:`ContigComposition` object consisting of the contiguous section.

    e.g.::

        ([2, 3],
         {'seq': 'chr', 'start': 3, 'end': 5, 'contig': 'contig1',
          'contig_start': 5, 'contig_end': 10, 'strand': '+',
          'contig_size': 20})

    indicates that we may wish to replace ``contig_composition_list[2:3]`` with
    the new :class:`ContigComposition` specified.

    Args:
        contig_composition_list (list): A list of :class:`ContigComposition`
            objects.
        verbose (bool, optional): If true, output additional debug info
            (default is False).

    Returns:
        list: A list of tuples with the indices of :class:`ContigComposition`
            objects in the list to be replaced along with replacement
            :class:`ContigComposition` objects.
    """
    previous_contigs = {}
    replacements = []
    for idx, section in enumerate(contig_composition_list):
        if verbose:
            section.to_tab()
        if section['contig'] in previous_contigs:
            if section['strand'] == previous_contigs[
                    section['contig']]['strand']:
                # Append to deletions and insertions
                if section['strand'] == '+':
                    start = previous_contigs[section['contig']]['end']
                    end = section['start']
                    contig_start = previous_contigs[
                        section['contig']]['contig_end']
                    contig_end = section['contig_start']
                    contig_size = section['contig_size']
                elif section['strand'] == '-':
                    start = previous_contigs[section['contig']]['end']
                    end = section['start']
                    contig_start = previous_contigs[
                        section['contig']]['contig_end']
                    contig_end = section['contig_start']
                    contig_size = section['contig_size']
                if contig_start != contig_end:
                    replacement = ContigComposition(
                        {'seq': section['seq'],
                         'start': start,
                         'end': end,
                         'contig': section['contig'],
                         'contig_start': contig_start,
                         'contig_end': contig_end,
                         'strand': section['strand'],
                         'contig_size': contig_size})
                    to_be_replaced = range(
                        previous_contigs[section['contig']]['last_index'] + 1,
                        idx)
                    replacements.append((to_be_replaced, replacement))
        previous_contigs[section['contig']] = section
        previous_contigs[section['contig']]['last_index'] = idx
    return (replacements)


def update_sequence_with_replacements(seq, replacements, replacement_seq_dict):
    """
    Update Seq object with replacements.

    The replacements specified by :class:`ContigComposition` objects and must
    be non-overlapping and sorted.

    Args:
        seq (Bio.Seq): The sequence object to be updated.
        replacements (list): A list :class:`ContigComposition` objects
        replacement_seq_dict (dict): A dictionary to the replacement sequences.

    Returns:
        Bio.Seq: An updated sequence object with replacements made
    """
    offset = 0
    updated_sequence = seq
    for replacement in replacements:
        (dummy_species, contig) = replacement['contig'].split('.', 1)
        contig_start = int(replacement['contig_start'])
        contig_end = int(replacement['contig_end'])
        contig_size = int(replacement['contig_size'])
        if replacement['strand'] == '+':
            replacement_seq = replacement_seq_dict[
                contig][contig_start:contig_end]
        else:
            rev_start = contig_size - contig_end
            rev_end = contig_size - contig_start
            replacement_seq = replacement_seq_dict[
                contig].reverse_complement()[rev_start:rev_end]
        start = int(replacement['start']) + offset
        end = int(replacement['end']) + offset
        updated_sequence = (updated_sequence[0:start] +
                            replacement_seq +
                            updated_sequence[end:len(seq)])
        updated_sequence.id = seq.id
        original_length = int(replacement['end']) - int(replacement['start'])
        replacement_length = contig_end - contig_start
        offset = offset + (replacement_length - original_length)
    return updated_sequence


def update_contig_composition_summary(contig_composition_summary,
                                      replacements):
    """
    Update list of :class:`ContigComposition` objects with replacements.

    Replacements are a list of tuples containing a list of indices of contigs
    to be replaced along with replacements.  The replacements must be
    non-overlapping and sorted.

    Args:
        contig_compostion_summary (list): List of dictionaries as returned by
            :meth:`summarize_contig_composition`
        replacements (list): A list of :class:`ContigCompostion` objects

    Retuns:
        list: A updated contig composition summary

    """

    last_cc_idx = 0
    updated_summary = []
    offset = 0
    for (replacement_indicies, replacement) in replacements:
        if len(replacement_indicies) >= 1:
            for idx in xrange(last_cc_idx, replacement_indicies[0]):
                updated_contig_composition = copy.deepcopy(
                    contig_composition_summary[idx])
                updated_contig_composition['start'] += offset
                updated_contig_composition['end'] += offset
                updated_summary.append(updated_contig_composition)
            last_cc_idx = replacement_indicies[-1] + 1
        else:
            for idx in xrange(last_cc_idx, len(contig_composition_summary)):
                if (contig_composition_summary[idx]['start'] <
                        replacement['start']):
                    updated_contig_composition = copy.deepcopy(
                        contig_composition_summary[idx])
                    updated_contig_composition['start'] += offset
                    updated_contig_composition['end'] += offset
                    updated_summary.append(updated_contig_composition)
                else:
                    last_cc_idx += 1
                    break
        updated_contig_composition = copy.deepcopy(replacement)
        updated_contig_composition['start'] += offset

        original_length = int(replacement['end']) - int(replacement['start'])
        replacement_length = (int(replacement['contig_end']) -
                              int(replacement['contig_start']))
        updated_contig_composition['end'] = (
            updated_contig_composition['start'] + replacement_length)
        updated_summary.append(updated_contig_composition)
        offset += replacement_length - original_length

    for idx in xrange(last_cc_idx, len(contig_composition_summary)):
        updated_contig_composition = copy.deepcopy(
            contig_composition_summary[idx])
        updated_contig_composition['start'] += offset
        updated_contig_composition['end'] += offset
        updated_summary.append(updated_contig_composition)
    return updated_summary


class ContigComposition(dict):
    """
    Represents the composition of one interval by another interval.

    Association of two genomic intervals, used to represent the composition
    of one interval by another.

    Attributes:
        seq (str): Sequence id of the interval being described
        start (str): The start position of the interval being described
            (1-based)
        end (str): The end position of the interval being described (1-based)
        contig (str): The sequence id of the second interval
        contig_start (str): The start position of the second interval (1-based)
        contig_end (str): The end position of the second interval (1-based)
        strand (str): The strand ('+' or '-') of the interval being described
        contig_size (str): The complete length of the sequence of the second
            interval.
    """

    def to_tab(self):
        """
        Return a tab delimited string of the ContigComposition

        Returns:
            string: A tab delimited string representing the ContigComposition.
        """
        return "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s" % (self.get("seq"),
                                                   self.get("start"),
                                                   self.get('end'),
                                                   self.get('contig'),
                                                   self.get('contig_start'),
                                                   self.get('contig_end'),
                                                   self.get('strand'),
                                                   self.get('contig_size'))

    @staticmethod
    def tab_headings():
        """
        Static method that retuns a tab delimited string of header names

        Returns:
            string: A tab delimited string representing the headers in the
                order used by the :meth:`to_tab` method.
        """
        return "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s" % ('seq',
                                                   'start',
                                                   'end',
                                                   'contig',
                                                   'contig_start',
                                                   'contig_end',
                                                   'strand',
                                                   'contig_size')
